@@ -968,40 +968,6 @@ asBignumObj(PyObject *value)
     return result;
 }
 
-static Tcl_Obj* AsObj(PyObject *value);
-
-static Tcl_Obj*
-TupleAsObj(PyObject *value, int wrapped)
-{
-    Tcl_Obj *result = NULL;
-    Py_ssize_t size = PyTuple_GET_SIZE(value);
-    if (size == 0) {
-        return Tcl_NewListObj(0, NULL);
-    }
-    if (!CHECK_SIZE(size, sizeof(Tcl_Obj *))) {
-        PyErr_SetString(PyExc_OverflowError,
-                        wrapped ? "list is too long" : "tuple is too long");
-        return NULL;
-    }
-    Tcl_Obj **argv = (Tcl_Obj **)PyMem_Malloc(((size_t)size) * sizeof(Tcl_Obj *));
-    if (argv == NULL) {
-      PyErr_NoMemory();
-      return NULL;
-    }
-    for (Py_ssize_t i = 0; i < size; i++) {
-        Tcl_Obj *item = AsObj(PyTuple_GET_ITEM(value, i));
-        if (item == NULL) {
-            goto exit;
-        }
-        argv[i] = item;
-    }
-    result = Tcl_NewListObj((int)size, argv);
-
-exit:
-    PyMem_Free(argv);
-    return result;
-}
-
 static Tcl_Obj*
 AsObj(PyObject *value)
 {
@@ -1048,17 +1014,28 @@ AsObj(PyObject *value)
     if (PyFloat_Check(value))
         return Tcl_NewDoubleObj(PyFloat_AS_DOUBLE(value));
 
-    if (PyTuple_Check(value)) {
-        return TupleAsObj(value, false);
-    }
+    if (PyTuple_Check(value) || PyList_Check(value)) {
+        Tcl_Obj **argv;
+        Py_ssize_t size, i;
 
-    if (PyList_Check(value)) {
-        PyObject *value_as_tuple = PyList_AsTuple(value);
-        if (value_as_tuple == NULL) {
+        size = PySequence_Fast_GET_SIZE(value);
+        if (size == 0)
+            return Tcl_NewListObj(0, NULL);
+        if (!CHECK_SIZE(size, sizeof(Tcl_Obj *))) {
+            PyErr_SetString(PyExc_OverflowError,
+                            PyTuple_Check(value) ? "tuple is too long" :
+                                                   "list is too long");
             return NULL;
         }
-        result = TupleAsObj(value_as_tuple, true);
-        Py_DECREF(value_as_tuple);
+        argv = (Tcl_Obj **) PyMem_Malloc(((size_t)size) * sizeof(Tcl_Obj *));
+        if (!argv) {
+          PyErr_NoMemory();
+          return NULL;
+        }
+        for (i = 0; i < size; i++)
+          argv[i] = AsObj(PySequence_Fast_GET_ITEM(value,i));
+        result = Tcl_NewListObj((int)size, argv);
+        PyMem_Free(argv);
         return result;
     }
 
