@@ -91,6 +91,12 @@ my_getpagesize(void)
 #  define MAP_ANONYMOUS MAP_ANON
 #endif
 
+/*[clinic input]
+module mmap
+class mmap.mmap "mmap_object *" ""
+[clinic start generated code]*/
+/*[clinic end generated code: output=da39a3ee5e6b4b0d input=82a9f8a529905b9b]*/
+
 typedef enum
 {
     ACCESS_DEFAULT,
@@ -120,22 +126,47 @@ typedef struct {
 #ifdef UNIX
     int fd;
     _Bool trackfd;
+    int flags;
 #endif
 
     PyObject *weakreflist;
     access_mode access;
 } mmap_object;
 
+#define mmap_object_CAST(op)    ((mmap_object *)(op))
+
+#include "clinic/mmapmodule.c.h"
+
+
+/* Return a Py_ssize_t from the object arg. This conversion logic is similar
+   to what AC uses for `Py_ssize_t` arguments.
+
+   Returns -1 on error. Use PyErr_Occurred() to disambiguate.
+*/
+static Py_ssize_t
+_As_Py_ssize_t(PyObject *arg) {
+    assert(arg != NULL);
+    Py_ssize_t ival = -1;
+    PyObject *iobj = _PyNumber_Index(arg);
+    if (iobj != NULL) {
+        ival = PyLong_AsSsize_t(iobj);
+        Py_DECREF(iobj);
+    }
+    return ival;
+}
+
+
 static int
-mmap_object_traverse(mmap_object *m_obj, visitproc visit, void *arg)
+mmap_object_traverse(PyObject *op, visitproc visit, void *arg)
 {
-    Py_VISIT(Py_TYPE(m_obj));
+    Py_VISIT(Py_TYPE(op));
     return 0;
 }
 
 static void
-mmap_object_dealloc(mmap_object *m_obj)
+mmap_object_dealloc(PyObject *op)
 {
+    mmap_object *m_obj = mmap_object_CAST(op);
     PyTypeObject *tp = Py_TYPE(m_obj);
     PyObject_GC_UnTrack(m_obj);
 
@@ -162,14 +193,21 @@ mmap_object_dealloc(mmap_object *m_obj)
     Py_END_ALLOW_THREADS
 #endif /* UNIX */
 
-    FT_CLEAR_WEAKREFS((PyObject *) m_obj, m_obj->weakreflist);
+    FT_CLEAR_WEAKREFS(op, m_obj->weakreflist);
 
     tp->tp_free(m_obj);
     Py_DECREF(tp);
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.close
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_close_method(mmap_object *self, PyObject *Py_UNUSED(ignored))
+mmap_mmap_close_impl(mmap_object *self)
+/*[clinic end generated code: output=a1ae0c727546f78d input=25020035f047eae1]*/
 {
     if (self->exports > 0) {
         PyErr_SetString(PyExc_BufferError, "cannot close "\
@@ -458,9 +496,15 @@ _safe_PyBytes_FromStringAndSize(char *start, size_t num_bytes) {
     }
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.read_byte
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_read_byte_method(mmap_object *self,
-                      PyObject *Py_UNUSED(ignored))
+mmap_mmap_read_byte_impl(mmap_object *self)
+/*[clinic end generated code: output=d931da1319f3869b input=5b8c6a904bdddda9]*/
 {
     CHECK_VALID(NULL);
     if (self->pos >= self->size) {
@@ -475,9 +519,15 @@ mmap_read_byte_method(mmap_object *self,
     return PyLong_FromLong((unsigned char) dest);
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.readline
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_read_line_method(mmap_object *self,
-                      PyObject *Py_UNUSED(ignored))
+mmap_mmap_readline_impl(mmap_object *self)
+/*[clinic end generated code: output=b9d2bf9999283311 input=2c4efd1d06e1cdd1]*/
 {
     Py_ssize_t remaining;
     char *start, *eol;
@@ -486,7 +536,7 @@ mmap_read_line_method(mmap_object *self,
 
     remaining = (self->pos < self->size) ? self->size - self->pos : 0;
     if (!remaining)
-        return PyBytes_FromString("");
+        return Py_GetConstant(Py_CONSTANT_EMPTY_BYTES);
     start = self->data + self->pos;
 
     if (safe_memchr(&eol, start, '\n', remaining) < 0) {
@@ -505,15 +555,21 @@ mmap_read_line_method(mmap_object *self,
     return result;
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.read
+
+  n as num_bytes: object(converter='_Py_convert_optional_to_ssize_t', type='Py_ssize_t', c_default='PY_SSIZE_T_MAX') = None
+  /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_read_method(mmap_object *self,
-                 PyObject *args)
+mmap_mmap_read_impl(mmap_object *self, Py_ssize_t num_bytes)
+/*[clinic end generated code: output=3b4d4f3704ed0969 input=8f97f361d435e357]*/
 {
-    Py_ssize_t num_bytes = PY_SSIZE_T_MAX, remaining;
+    Py_ssize_t remaining;
 
-    CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "|O&:read", _Py_convert_optional_to_ssize_t, &num_bytes))
-        return NULL;
     CHECK_VALID(NULL);
 
     /* silently 'adjust' out-of-range requests */
@@ -530,81 +586,105 @@ mmap_read_method(mmap_object *self,
 }
 
 static PyObject *
-mmap_gfind(mmap_object *self,
-           PyObject *args,
-           int reverse)
+mmap_gfind_lock_held(mmap_object *self, Py_buffer *view, PyObject *start_obj,
+                    PyObject *end_obj, int reverse)
 {
     Py_ssize_t start = self->pos;
     Py_ssize_t end = self->size;
-    Py_buffer view;
 
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, reverse ? "y*|nn:rfind" : "y*|nn:find",
-                          &view, &start, &end)) {
-        return NULL;
-    }
-    else {
-        if (start < 0)
-            start += self->size;
-        if (start < 0)
-            start = 0;
-        else if (start > self->size)
-            start = self->size;
-
-        if (end < 0)
-            end += self->size;
-        if (end < 0)
-            end = 0;
-        else if (end > self->size)
-            end = self->size;
-
-        Py_ssize_t index;
-        PyObject *result;
-        CHECK_VALID_OR_RELEASE(NULL, view);
-        if (end < start) {
-            result = PyLong_FromSsize_t(-1);
+    if (start_obj != Py_None) {
+        start = _As_Py_ssize_t(start_obj);
+        if (start == -1 && PyErr_Occurred()) {
+            return NULL;
         }
-        else if (reverse) {
-            assert(0 <= start && start <= end && end <= self->size);
-            if (_safe_PyBytes_ReverseFind(&index, self,
-                self->data + start, end - start,
-                view.buf, view.len, start) < 0)
-            {
-                result = NULL;
-            }
-            else {
-                result = PyLong_FromSsize_t(index);
+
+        if (end_obj != Py_None) {
+            end = _As_Py_ssize_t(end_obj);
+            if (end == -1 && PyErr_Occurred()) {
+                return NULL;
             }
         }
+    }
+
+    if (start < 0)
+        start += self->size;
+    if (start < 0)
+        start = 0;
+    else if (start > self->size)
+        start = self->size;
+
+    if (end < 0)
+        end += self->size;
+    if (end < 0)
+        end = 0;
+    else if (end > self->size)
+        end = self->size;
+
+    Py_ssize_t index;
+    PyObject *result;
+    CHECK_VALID(NULL);
+    if (end < start) {
+        result = PyLong_FromSsize_t(-1);
+    }
+    else if (reverse) {
+        assert(0 <= start && start <= end && end <= self->size);
+        if (_safe_PyBytes_ReverseFind(&index, self,
+            self->data + start, end - start,
+            view->buf, view->len, start) < 0)
+        {
+            result = NULL;
+        }
         else {
-            assert(0 <= start && start <= end && end <= self->size);
-            if (_safe_PyBytes_Find(&index, self,
-                self->data + start, end - start,
-                view.buf, view.len, start) < 0)
-            {
-                result = NULL;
-            }
-            else {
-                result = PyLong_FromSsize_t(index);
-            }
+            result = PyLong_FromSsize_t(index);
         }
-        PyBuffer_Release(&view);
-        return result;
     }
+    else {
+        assert(0 <= start && start <= end && end <= self->size);
+        if (_safe_PyBytes_Find(&index, self,
+            self->data + start, end - start,
+            view->buf, view->len, start) < 0)
+        {
+            result = NULL;
+        }
+        else {
+            result = PyLong_FromSsize_t(index);
+        }
+    }
+    return result;
 }
 
-static PyObject *
-mmap_find_method(mmap_object *self,
-                 PyObject *args)
-{
-    return mmap_gfind(self, args, 0);
-}
+/*[clinic input]
+@critical_section
+mmap.mmap.find
+
+  view: Py_buffer
+  start: object = None
+  end: object = None
+  /
+
+[clinic start generated code]*/
 
 static PyObject *
-mmap_rfind_method(mmap_object *self,
-                 PyObject *args)
+mmap_mmap_find_impl(mmap_object *self, Py_buffer *view, PyObject *start,
+                    PyObject *end)
+/*[clinic end generated code: output=ef8878a322f00192 input=0135504494b52c2b]*/
 {
-    return mmap_gfind(self, args, 1);
+    return mmap_gfind_lock_held(self, view, start, end, 0);
+}
+
+/*[clinic input]
+@critical_section
+mmap.mmap.rfind = mmap.mmap.find
+
+[clinic start generated code]*/
+
+static PyObject *
+mmap_mmap_rfind_impl(mmap_object *self, Py_buffer *view, PyObject *start,
+                     PyObject *end)
+/*[clinic end generated code: output=73b918940d67c2b8 input=8aecdd1f70c06c62]*/
+{
+    return mmap_gfind_lock_held(self, view, start, end, 1);
 }
 
 static int
@@ -640,50 +720,55 @@ is_resizeable(mmap_object *self)
 }
 
 
+/*[clinic input]
+@critical_section
+mmap.mmap.write
+
+    bytes as data: Py_buffer
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_write_method(mmap_object *self,
-                  PyObject *args)
+mmap_mmap_write_impl(mmap_object *self, Py_buffer *data)
+/*[clinic end generated code: output=9e97063efb6fb27b input=3f16fa79aa89d6f7]*/
 {
-    Py_buffer data;
-
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "y*:write", &data))
-        return NULL;
-
     if (!is_writable(self)) {
-        PyBuffer_Release(&data);
         return NULL;
     }
 
-    if (self->pos > self->size || self->size - self->pos < data.len) {
-        PyBuffer_Release(&data);
+    if (self->pos > self->size || self->size - self->pos < data->len) {
         PyErr_SetString(PyExc_ValueError, "data out of range");
         return NULL;
     }
 
-    CHECK_VALID_OR_RELEASE(NULL, data);
+    CHECK_VALID(NULL);
     PyObject *result;
-    if (safe_memcpy(self->data + self->pos, data.buf, data.len) < 0) {
+    if (safe_memcpy(self->data + self->pos, data->buf, data->len) < 0) {
         result = NULL;
     }
     else {
-        self->pos += data.len;
-        result = PyLong_FromSsize_t(data.len);
+        self->pos += data->len;
+        result = PyLong_FromSsize_t(data->len);
     }
-    PyBuffer_Release(&data);
     return result;
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.write_byte
+
+    byte as value: unsigned_char
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_write_byte_method(mmap_object *self,
-                       PyObject *args)
+mmap_mmap_write_byte_impl(mmap_object *self, unsigned char value)
+/*[clinic end generated code: output=aa11adada9b17510 input=32740bfa174f0991]*/
 {
-    char value;
-
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "b:write_byte", &value))
-        return(NULL);
-
     if (!is_writable(self))
         return NULL;
 
@@ -693,16 +778,22 @@ mmap_write_byte_method(mmap_object *self,
         return NULL;
     }
 
-    if (safe_byte_copy(self->data + self->pos, &value) < 0) {
+    if (safe_byte_copy(self->data + self->pos, (const char*)&value) < 0) {
         return NULL;
     }
     self->pos++;
     Py_RETURN_NONE;
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.size
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_size_method(mmap_object *self,
-                 PyObject *Py_UNUSED(ignored))
+mmap_mmap_size_impl(mmap_object *self)
+/*[clinic end generated code: output=c177e65e83a648ff input=f69c072efd2e1595]*/
 {
     CHECK_VALID(NULL);
 
@@ -750,14 +841,21 @@ mmap_size_method(mmap_object *self,
  / new size?
  */
 
+/*[clinic input]
+@critical_section
+mmap.mmap.resize
+
+    newsize as new_size: Py_ssize_t
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_resize_method(mmap_object *self,
-                   PyObject *args)
+mmap_mmap_resize_impl(mmap_object *self, Py_ssize_t new_size)
+/*[clinic end generated code: output=6f262537ce9c2dcc input=b6b5dee52a41b79f]*/
 {
-    Py_ssize_t new_size;
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "n:resize", &new_size) ||
-        !is_resizeable(self)) {
+    if (!is_resizeable(self)) {
         return NULL;
     }
     if (new_size < 0 || PY_SSIZE_T_MAX - new_size < self->offset) {
@@ -871,6 +969,13 @@ mmap_resize_method(mmap_object *self,
 #else
         void *newmap;
 
+#ifdef __linux__
+        if (self->fd == -1 && !(self->flags & MAP_PRIVATE) && new_size > self->size) {
+            PyErr_Format(PyExc_ValueError,
+                "mmap: can't expand a shared anonymous mapping on Linux");
+            return NULL;
+        }
+#endif
         if (self->fd != -1 && ftruncate(self->fd, self->offset + new_size) == -1) {
             PyErr_SetFromErrno(PyExc_OSError);
             return NULL;
@@ -898,21 +1003,45 @@ mmap_resize_method(mmap_object *self,
     }
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.tell
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_tell_method(mmap_object *self, PyObject *Py_UNUSED(ignored))
+mmap_mmap_tell_impl(mmap_object *self)
+/*[clinic end generated code: output=6034958630e1b1d1 input=fd163acacf45c3a5]*/
 {
     CHECK_VALID(NULL);
     return PyLong_FromSize_t(self->pos);
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.flush
+
+    offset: Py_ssize_t = 0
+    size as size_obj: object = None
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_flush_method(mmap_object *self, PyObject *args)
+mmap_mmap_flush_impl(mmap_object *self, Py_ssize_t offset,
+                     PyObject *size_obj)
+/*[clinic end generated code: output=41a10c349ed1608a input=7e6bb8f9462f53e6]*/
 {
-    Py_ssize_t offset = 0;
     Py_ssize_t size = self->size;
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "|nn:flush", &offset, &size))
-        return NULL;
+
+    if (size_obj != Py_None) {
+        size = _As_Py_ssize_t(size_obj);
+        if (size == -1 && PyErr_Occurred()) {
+            return NULL;
+        }
+    }
+
     if (size < 0 || offset < 0 || self->size - offset < size) {
         PyErr_SetString(PyExc_ValueError, "flush values out of range");
         return NULL;
@@ -940,58 +1069,80 @@ mmap_flush_method(mmap_object *self, PyObject *args)
 #endif
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.seek
+
+    pos as dist: Py_ssize_t
+    whence as how: int = 0
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_seek_method(mmap_object *self, PyObject *args)
+mmap_mmap_seek_impl(mmap_object *self, Py_ssize_t dist, int how)
+/*[clinic end generated code: output=00310494e8b8c592 input=e2fda5d081c3db22]*/
 {
-    Py_ssize_t dist;
-    int how=0;
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "n|i:seek", &dist, &how))
-        return NULL;
-    else {
-        Py_ssize_t where;
-        switch (how) {
-        case 0: /* relative to start */
-            where = dist;
-            break;
-        case 1: /* relative to current position */
-            if (PY_SSIZE_T_MAX - self->pos < dist)
-                goto onoutofrange;
-            where = self->pos + dist;
-            break;
-        case 2: /* relative to end */
-            if (PY_SSIZE_T_MAX - self->size < dist)
-                goto onoutofrange;
-            where = self->size + dist;
-            break;
-        default:
-            PyErr_SetString(PyExc_ValueError, "unknown seek type");
-            return NULL;
-        }
-        if (where > self->size || where < 0)
+    Py_ssize_t where;
+    switch (how) {
+    case 0: /* relative to start */
+        where = dist;
+        break;
+    case 1: /* relative to current position */
+        if (PY_SSIZE_T_MAX - self->pos < dist)
             goto onoutofrange;
-        self->pos = where;
-        return PyLong_FromSsize_t(self->pos);
+        where = self->pos + dist;
+        break;
+    case 2: /* relative to end */
+        if (PY_SSIZE_T_MAX - self->size < dist)
+            goto onoutofrange;
+        where = self->size + dist;
+        break;
+    default:
+        PyErr_SetString(PyExc_ValueError, "unknown seek type");
+        return NULL;
     }
+    if (where > self->size || where < 0)
+        goto onoutofrange;
+    self->pos = where;
+    return PyLong_FromSsize_t(self->pos);
 
   onoutofrange:
     PyErr_SetString(PyExc_ValueError, "seek out of range");
     return NULL;
 }
 
+/*[clinic input]
+mmap.mmap.seekable
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_seekable_method(mmap_object *self, PyObject *Py_UNUSED(ignored))
+mmap_mmap_seekable_impl(mmap_object *self)
+/*[clinic end generated code: output=6311dc3ea300fa38 input=5132505f6e259001]*/
 {
     Py_RETURN_TRUE;
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.move
+
+    dest: Py_ssize_t
+    src: Py_ssize_t
+    count as cnt: Py_ssize_t
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_move_method(mmap_object *self, PyObject *args)
+mmap_mmap_move_impl(mmap_object *self, Py_ssize_t dest, Py_ssize_t src,
+                    Py_ssize_t cnt)
+/*[clinic end generated code: output=391f549a44181793 input=cf8cfe10d9f6b448]*/
 {
-    Py_ssize_t dest, src, cnt;
     CHECK_VALID(NULL);
-    if (!PyArg_ParseTuple(args, "nnn:move", &dest, &src, &cnt) ||
-        !is_writable(self)) {
+    if (!is_writable(self)) {
         return NULL;
     } else {
         /* bounds check the values */
@@ -1014,33 +1165,58 @@ mmap_move_method(mmap_object *self, PyObject *args)
 }
 
 static PyObject *
-mmap_closed_get(mmap_object *self, void *Py_UNUSED(ignored))
+mmap_closed_get(PyObject *op, void *Py_UNUSED(closure))
 {
+    mmap_object *self = mmap_object_CAST(op);
+    PyObject *result;
+    Py_BEGIN_CRITICAL_SECTION(op);
 #ifdef MS_WINDOWS
-    return PyBool_FromLong(self->map_handle == NULL ? 1 : 0);
+    result = PyBool_FromLong(self->map_handle == NULL ? 1 : 0);
 #elif defined(UNIX)
-    return PyBool_FromLong(self->data == NULL ? 1 : 0);
+    result = PyBool_FromLong(self->data == NULL ? 1 : 0);
 #endif
+    Py_END_CRITICAL_SECTION();
+    return result;
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.__enter__
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap__enter__method(mmap_object *self, PyObject *args)
+mmap_mmap___enter___impl(mmap_object *self)
+/*[clinic end generated code: output=92cfc59f4c4e2d26 input=a446541fbfe0b890]*/
 {
     CHECK_VALID(NULL);
 
     return Py_NewRef(self);
 }
 
+/*[clinic input]
+@critical_section
+mmap.mmap.__exit__
+
+    exc_type: object
+    exc_value: object
+    traceback: object
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap__exit__method(PyObject *self, PyObject *args)
+mmap_mmap___exit___impl(mmap_object *self, PyObject *exc_type,
+                        PyObject *exc_value, PyObject *traceback)
+/*[clinic end generated code: output=bec7e3e319c1f07e input=5f28e91cf752bc64]*/
 {
-    return mmap_close_method((mmap_object *)self, NULL);
+    return mmap_mmap_close_impl(self);
 }
 
 static PyObject *
-mmap__repr__method(PyObject *self)
+mmap__repr__method_lock_held(PyObject *op)
 {
-    mmap_object *mobj = (mmap_object *)self;
+    mmap_object *mobj = mmap_object_CAST(op);
 
 #ifdef MS_WINDOWS
 #define _Py_FORMAT_OFFSET "lld"
@@ -1054,7 +1230,7 @@ mmap__repr__method(PyObject *self)
     if (mobj->data == NULL)
 #endif
     {
-        return PyUnicode_FromFormat("<%s closed=True>", Py_TYPE(self)->tp_name);
+        return PyUnicode_FromFormat("<%s closed=True>", Py_TYPE(op)->tp_name);
     } else {
         const char *access_str;
 
@@ -1077,14 +1253,31 @@ mmap__repr__method(PyObject *self)
 
         return PyUnicode_FromFormat("<%s closed=False, access=%s, length=%zd, "
                                     "pos=%zd, offset=%" _Py_FORMAT_OFFSET ">",
-                                    Py_TYPE(self)->tp_name, access_str,
+                                    Py_TYPE(op)->tp_name, access_str,
                                     mobj->size, mobj->pos, mobj->offset);
     }
 }
 
-#ifdef MS_WINDOWS
 static PyObject *
-mmap__sizeof__method(mmap_object *self, void *Py_UNUSED(ignored))
+mmap__repr__method(PyObject *op)
+{
+    PyObject *result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap__repr__method_lock_held(op);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
+#ifdef MS_WINDOWS
+/*[clinic input]
+@critical_section
+mmap.mmap.__sizeof__
+
+[clinic start generated code]*/
+
+static PyObject *
+mmap_mmap___sizeof___impl(mmap_object *self)
+/*[clinic end generated code: output=1aed30daff807d09 input=8a648868a089553c]*/
 {
     size_t res = _PyObject_SIZE(Py_TYPE(self));
     if (self->tagname) {
@@ -1095,17 +1288,26 @@ mmap__sizeof__method(mmap_object *self, void *Py_UNUSED(ignored))
 #endif
 
 #if defined(MS_WINDOWS) && defined(Py_DEBUG)
+/*[clinic input]
+@critical_section
+mmap.mmap._protect
+
+    flNewProtect: unsigned_int(bitwise=True)
+    start: Py_ssize_t
+    length: Py_ssize_t
+    /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_protect_method(mmap_object *self, PyObject *args) {
-    DWORD flNewProtect, flOldProtect;
-    Py_ssize_t start, length;
+mmap_mmap__protect_impl(mmap_object *self, unsigned int flNewProtect,
+                        Py_ssize_t start, Py_ssize_t length)
+/*[clinic end generated code: output=a87271a34d1ad6cf input=9170498c5e1482da]*/
+{
+    DWORD flOldProtect;
 
     CHECK_VALID(NULL);
 
-    if (!PyArg_ParseTuple(args, "Inn:protect", &flNewProtect, &start, &length)) {
-        return NULL;
-    }
-
     if (!VirtualProtect((void *) (self->data + start), length, flNewProtect,
                         &flOldProtect))
     {
@@ -1118,17 +1320,32 @@ mmap_protect_method(mmap_object *self, PyObject *args) {
 #endif
 
 #ifdef HAVE_MADVISE
+/*[clinic input]
+@critical_section
+mmap.mmap.madvise
+
+  option: int
+  start: Py_ssize_t = 0
+  length as length_obj: object = None
+  /
+
+[clinic start generated code]*/
+
 static PyObject *
-mmap_madvise_method(mmap_object *self, PyObject *args)
+mmap_mmap_madvise_impl(mmap_object *self, int option, Py_ssize_t start,
+                       PyObject *length_obj)
+/*[clinic end generated code: output=816be656f08c0e3c input=2d37f7a4c87f1053]*/
 {
-    int option;
-    Py_ssize_t start = 0, length;
+    Py_ssize_t length;
 
     CHECK_VALID(NULL);
-    length = self->size;
-
-    if (!PyArg_ParseTuple(args, "i|nn:madvise", &option, &start, &length)) {
-        return NULL;
+    if (length_obj == Py_None) {
+        length = self->size;
+    } else {
+        length = _As_Py_ssize_t(length_obj);
+        if (length == -1 && PyErr_Occurred()) {
+            return NULL;
+        }
     }
 
     if (start < 0 || start >= self->size) {
@@ -1164,37 +1381,31 @@ static struct PyMemberDef mmap_object_members[] = {
 };
 
 static struct PyMethodDef mmap_object_methods[] = {
-    {"close",           (PyCFunction) mmap_close_method,        METH_NOARGS},
-    {"find",            (PyCFunction) mmap_find_method,         METH_VARARGS},
-    {"rfind",           (PyCFunction) mmap_rfind_method,        METH_VARARGS},
-    {"flush",           (PyCFunction) mmap_flush_method,        METH_VARARGS},
-#ifdef HAVE_MADVISE
-    {"madvise",         (PyCFunction) mmap_madvise_method,      METH_VARARGS},
-#endif
-    {"move",            (PyCFunction) mmap_move_method,         METH_VARARGS},
-    {"read",            (PyCFunction) mmap_read_method,         METH_VARARGS},
-    {"read_byte",       (PyCFunction) mmap_read_byte_method,    METH_NOARGS},
-    {"readline",        (PyCFunction) mmap_read_line_method,    METH_NOARGS},
-    {"resize",          (PyCFunction) mmap_resize_method,       METH_VARARGS},
-    {"seek",            (PyCFunction) mmap_seek_method,         METH_VARARGS},
-    {"seekable",        (PyCFunction) mmap_seekable_method,     METH_NOARGS},
-    {"size",            (PyCFunction) mmap_size_method,         METH_NOARGS},
-    {"tell",            (PyCFunction) mmap_tell_method,         METH_NOARGS},
-    {"write",           (PyCFunction) mmap_write_method,        METH_VARARGS},
-    {"write_byte",      (PyCFunction) mmap_write_byte_method,   METH_VARARGS},
-    {"__enter__",       (PyCFunction) mmap__enter__method,      METH_NOARGS},
-    {"__exit__",        (PyCFunction) mmap__exit__method,       METH_VARARGS},
-#ifdef MS_WINDOWS
-    {"__sizeof__",      (PyCFunction) mmap__sizeof__method,     METH_NOARGS},
-#ifdef Py_DEBUG
-    {"_protect",        (PyCFunction) mmap_protect_method,      METH_VARARGS},
-#endif // Py_DEBUG
-#endif // MS_WINDOWS
+    MMAP_MMAP_CLOSE_METHODDEF
+    MMAP_MMAP_FIND_METHODDEF
+    MMAP_MMAP_RFIND_METHODDEF
+    MMAP_MMAP_FLUSH_METHODDEF
+    MMAP_MMAP_MADVISE_METHODDEF
+    MMAP_MMAP_MOVE_METHODDEF
+    MMAP_MMAP_READ_METHODDEF
+    MMAP_MMAP_READ_BYTE_METHODDEF
+    MMAP_MMAP_READLINE_METHODDEF
+    MMAP_MMAP_RESIZE_METHODDEF
+    MMAP_MMAP_SEEK_METHODDEF
+    MMAP_MMAP_SEEKABLE_METHODDEF
+    MMAP_MMAP_SIZE_METHODDEF
+    MMAP_MMAP_TELL_METHODDEF
+    MMAP_MMAP_WRITE_METHODDEF
+    MMAP_MMAP_WRITE_BYTE_METHODDEF
+    MMAP_MMAP___ENTER___METHODDEF
+    MMAP_MMAP___EXIT___METHODDEF
+    MMAP_MMAP___SIZEOF___METHODDEF
+    MMAP_MMAP__PROTECT_METHODDEF
     {NULL,         NULL}       /* sentinel */
 };
 
 static PyGetSetDef mmap_object_getset[] = {
-    {"closed", (getter) mmap_closed_get, NULL, NULL},
+    {"closed", mmap_closed_get, NULL, NULL},
     {NULL}
 };
 
@@ -1202,32 +1413,58 @@ static PyGetSetDef mmap_object_getset[] = {
 /* Functions for treating an mmap'ed file as a buffer */
 
 static int
-mmap_buffer_getbuf(mmap_object *self, Py_buffer *view, int flags)
+mmap_buffer_getbuf_lock_held(PyObject *op, Py_buffer *view, int flags)
 {
+    mmap_object *self = mmap_object_CAST(op);
     CHECK_VALID(-1);
-    if (PyBuffer_FillInfo(view, (PyObject*)self, self->data, self->size,
+    if (PyBuffer_FillInfo(view, op, self->data, self->size,
                           (self->access == ACCESS_READ), flags) < 0)
         return -1;
     self->exports++;
     return 0;
 }
 
-static void
-mmap_buffer_releasebuf(mmap_object *self, Py_buffer *view)
+static int
+mmap_buffer_getbuf(PyObject *op, Py_buffer *view, int flags)
 {
+    int result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap_buffer_getbuf_lock_held(op, view, flags);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
+static void
+mmap_buffer_releasebuf(PyObject *op, Py_buffer *Py_UNUSED(view))
+{
+    mmap_object *self = mmap_object_CAST(op);
+    Py_BEGIN_CRITICAL_SECTION(self);
     self->exports--;
+    Py_END_CRITICAL_SECTION();
 }
 
 static Py_ssize_t
-mmap_length(mmap_object *self)
+mmap_length_lock_held(PyObject *op)
 {
+    mmap_object *self = mmap_object_CAST(op);
     CHECK_VALID(-1);
     return self->size;
 }
 
-static PyObject *
-mmap_item(mmap_object *self, Py_ssize_t i)
+static Py_ssize_t
+mmap_length(PyObject *op)
 {
+    Py_ssize_t result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap_length_lock_held(op);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
+static PyObject *
+mmap_item_lock_held(PyObject *op, Py_ssize_t i)
+{
+    mmap_object *self = mmap_object_CAST(op);
     CHECK_VALID(NULL);
     if (i < 0 || i >= self->size) {
         PyErr_SetString(PyExc_IndexError, "mmap index out of range");
@@ -1242,8 +1479,18 @@ mmap_item(mmap_object *self, Py_ssize_t i)
 }
 
 static PyObject *
-mmap_subscript(mmap_object *self, PyObject *item)
+mmap_item(PyObject *op, Py_ssize_t i) {
+    PyObject *result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap_item_lock_held(op, i);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
+static PyObject *
+mmap_subscript_lock_held(PyObject *op, PyObject *item)
 {
+    mmap_object *self = mmap_object_CAST(op);
     CHECK_VALID(NULL);
     if (PyIndex_Check(item)) {
         Py_ssize_t i = PyNumber_AsSsize_t(item, PyExc_IndexError);
@@ -1274,7 +1521,7 @@ mmap_subscript(mmap_object *self, PyObject *item)
 
         CHECK_VALID(NULL);
         if (slicelen <= 0)
-            return PyBytes_FromStringAndSize("", 0);
+            return Py_GetConstant(Py_CONSTANT_EMPTY_BYTES);
         else if (step == 1)
             return _safe_PyBytes_FromStringAndSize(self->data + start, slicelen);
         else {
@@ -1303,10 +1550,21 @@ mmap_subscript(mmap_object *self, PyObject *item)
     }
 }
 
+static PyObject *
+mmap_subscript(PyObject *op, PyObject *item)
+{
+    PyObject *result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap_subscript_lock_held(op, item);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
 static int
-mmap_ass_item(mmap_object *self, Py_ssize_t i, PyObject *v)
+mmap_ass_item_lock_held(PyObject *op, Py_ssize_t i, PyObject *v)
 {
     const char *buf;
+    mmap_object *self = mmap_object_CAST(op);
 
     CHECK_VALID(-1);
     if (i < 0 || i >= self->size) {
@@ -1334,8 +1592,19 @@ mmap_ass_item(mmap_object *self, Py_ssize_t i, PyObject *v)
 }
 
 static int
-mmap_ass_subscript(mmap_object *self, PyObject *item, PyObject *value)
+mmap_ass_item(PyObject *op, Py_ssize_t i, PyObject *v)
 {
+    int result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap_ass_item_lock_held(op, i, v);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
+static int
+mmap_ass_subscript_lock_held(PyObject *op, PyObject *item, PyObject *value)
+{
+    mmap_object *self = mmap_object_CAST(op);
     CHECK_VALID(-1);
 
     if (!is_writable(self))
@@ -1429,6 +1698,16 @@ mmap_ass_subscript(mmap_object *self, PyObject *item, PyObject *value)
     }
 }
 
+static int
+mmap_ass_subscript(PyObject *op, PyObject *item, PyObject *value)
+{
+    int result;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    result = mmap_ass_subscript_lock_held(op, item, value);
+    Py_END_CRITICAL_SECTION();
+    return result;
+}
+
 static PyObject *
 new_mmap_object(PyTypeObject *type, PyObject *args, PyObject *kwdict);
 
@@ -1651,6 +1930,7 @@ new_mmap_object(PyTypeObject *type, PyObject *args, PyObject *kwdict)
     else {
         m_obj->fd = -1;
     }
+    m_obj->flags = flags;
 
     Py_BEGIN_ALLOW_THREADS
     m_obj->data = mmap(NULL, map_size, prot, flags, fd, offset);
