@@ -1806,7 +1806,7 @@ _overlapped_Overlapped_WSASendTo_impl(OverlappedObject *self, HANDLE handle,
         case ERROR_IO_PENDING:
             Py_RETURN_NONE;
         default:
-            Overlapped_clear(self);
+            self->type = TYPE_NOT_STARTED;
             return SetFromWindowsErr(err);
     }
 }
@@ -1873,7 +1873,7 @@ _overlapped_Overlapped_WSARecvFrom_impl(OverlappedObject *self,
     case ERROR_IO_PENDING:
         Py_RETURN_NONE;
     default:
-        Overlapped_clear(self);
+        self->type = TYPE_NOT_STARTED;
         return SetFromWindowsErr(err);
     }
 }
@@ -1940,7 +1940,7 @@ _overlapped_Overlapped_WSARecvFromInto_impl(OverlappedObject *self,
     case ERROR_IO_PENDING:
         Py_RETURN_NONE;
     default:
-        Overlapped_clear(self);
+        self->type = TYPE_NOT_STARTED;
         return SetFromWindowsErr(err);
     }
 }
