@@ -200,7 +200,7 @@ Here are some ways to address this challenge:
 Adding the checks to the concrete API would help make any interpreter
 switch to OrderedDict less painful for extension modules.  However, this
 won't work.  The equivalent C API call to `dict.__setitem__(obj, k, v)`
-is 'PyDict_SetItem(obj, k, v)`.  This illustrates how subclasses in C call
+is `PyDict_SetItem(obj, k, v)`.  This illustrates how subclasses in C call
 the base class's methods, since there is no equivalent of super() in the
 C API.  Calling into Python for parent class API would work, but some
 extension modules already rely on this feature of the concrete API.
@@ -260,7 +260,7 @@ mp_length         __len__          -                   dict_length
 mp_subscript      __getitem__      -                   dict_subscript
 mp_ass_subscript  __setitem__      -                   dict_ass_sub
                   __delitem__
-tp_hash           __hash__         _Py_HashPointer     ..._HashNotImpl
+tp_hash           __hash__         Py_HashPointer      ..._HashNotImpl
 tp_str            __str__          object_str          -
 tp_getattro       __getattribute__ ..._GenericGetAttr  (repeated)
                   __getattr__
@@ -276,7 +276,7 @@ tp_dictoffset     (__dict__)       -                   -
 tp_init           __init__         object_init         dict_init
 tp_alloc          -                PyType_GenericAlloc (repeated)
 tp_new            __new__          object_new          dict_new
-tp_free           -                PyObject_Del        PyObject_GC_Del
+tp_free           -                PyObject_Free       PyObject_GC_Del
 ================= ================ =================== ================
 
 Relevant Methods
@@ -471,6 +471,7 @@ later:
 #include "pycore_dict.h"             // _Py_dict_lookup()
 #include "pycore_object.h"           // _PyObject_GC_UNTRACK()
 #include "pycore_pyerrors.h"         // _PyErr_ChainExceptions1()
+#include "pycore_tuple.h"            // _PyTuple_Recycle()
 #include <stddef.h>                  // offsetof()
 #include "pycore_weakref.h"          // FT_CLEAR_WEAKREFS()
 
@@ -501,6 +502,8 @@ struct _odictobject {
     PyObject *od_weakreflist;    /* holds weakrefs to the odict */
 };
 
+#define _PyODictObject_CAST(op) _Py_CAST(PyODictObject*, (op))
+
 
 /* ----------------------------------------------
  * odict keys (a simple doubly-linked list)
@@ -523,8 +526,8 @@ struct _odictnode {
 #define _odictnode_PREV(node) (node->prev)
 #define _odictnode_NEXT(node) (node->next)
 
-#define _odict_FIRST(od) (((PyODictObject *)od)->od_first)
-#define _odict_LAST(od) (((PyODictObject *)od)->od_last)
+#define _odict_FIRST(od) (_PyODictObject_CAST(od)->od_first)
+#define _odict_LAST(od) (_PyODictObject_CAST(od)->od_last)
 #define _odict_EMPTY(od) (_odict_FIRST(od) == NULL)
 #define _odict_FOREACH(od, node) \
     for (node = _odict_FIRST(od); node != NULL; node = _odictnode_NEXT(node))
@@ -533,6 +536,7 @@ struct _odictnode {
 static Py_ssize_t
 _odict_get_index_raw(PyODictObject *od, PyObject *key, Py_hash_t hash)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     PyObject *value = NULL;
     PyDictKeysObject *keys = ((PyDictObject *)od)->ma_keys;
     Py_ssize_t ix;
@@ -557,6 +561,7 @@ _odict_get_index_raw(PyODictObject *od, PyObject *key, Py_hash_t hash)
 static int
 _odict_resize(PyODictObject *od)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     Py_ssize_t size, i;
     _ODictNode **fast_nodes, *node;
 
@@ -593,6 +598,7 @@ _odict_resize(PyODictObject *od)
 static Py_ssize_t
 _odict_get_index(PyODictObject *od, PyObject *key, Py_hash_t hash)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     PyDictKeysObject *keys;
 
     assert(key != NULL);
@@ -613,6 +619,7 @@ _odict_get_index(PyODictObject *od, PyObject *key, Py_hash_t hash)
 static _ODictNode *
 _odict_find_node_hash(PyODictObject *od, PyObject *key, Py_hash_t hash)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     Py_ssize_t index;
 
     if (_odict_EMPTY(od))
@@ -627,6 +634,7 @@ _odict_find_node_hash(PyODictObject *od, PyObject *key, Py_hash_t hash)
 static _ODictNode *
 _odict_find_node(PyODictObject *od, PyObject *key)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     Py_ssize_t index;
     Py_hash_t hash;
 
@@ -645,6 +653,7 @@ _odict_find_node(PyODictObject *od, PyObject *key)
 static void
 _odict_add_head(PyODictObject *od, _ODictNode *node)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     _odictnode_PREV(node) = NULL;
     _odictnode_NEXT(node) = _odict_FIRST(od);
     if (_odict_FIRST(od) == NULL)
@@ -658,6 +667,7 @@ _odict_add_head(PyODictObject *od, _ODictNode *node)
 static void
 _odict_add_tail(PyODictObject *od, _ODictNode *node)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     _odictnode_PREV(node) = _odict_LAST(od);
     _odictnode_NEXT(node) = NULL;
     if (_odict_LAST(od) == NULL)
@@ -672,6 +682,7 @@ _odict_add_tail(PyODictObject *od, _ODictNode *node)
 static int
 _odict_add_new_node(PyODictObject *od, PyObject *key, Py_hash_t hash)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     Py_ssize_t i;
     _ODictNode *node;
 
@@ -716,6 +727,7 @@ _odict_add_new_node(PyODictObject *od, PyObject *key, Py_hash_t hash)
 static void
 _odict_remove_node(PyODictObject *od, _ODictNode *node)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     if (_odict_FIRST(od) == node)
         _odict_FIRST(od) = _odictnode_NEXT(node);
     else if (_odictnode_PREV(node) != NULL)
@@ -751,6 +763,7 @@ static int
 _odict_clear_node(PyODictObject *od, _ODictNode *node, PyObject *key,
                   Py_hash_t hash)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
     Py_ssize_t i;
 
     assert(key != NULL);
@@ -858,12 +871,12 @@ _odict_keys_equal(PyODictObject *a, PyODictObject *b)
 /* mp_ass_subscript: __setitem__() and __delitem__() */
 
 static int
-odict_mp_ass_sub(PyODictObject *od, PyObject *v, PyObject *w)
+odict_mp_ass_sub(PyObject *od, PyObject *v, PyObject *w)
 {
     if (w == NULL)
-        return PyODict_DelItem((PyObject *)od, v);
+        return PyODict_DelItem(od, v);
     else
-        return PyODict_SetItem((PyObject *)od, v, w);
+        return PyODict_SetItem(od, v, w);
 }
 
 /* tp_as_mapping */
@@ -871,7 +884,7 @@ odict_mp_ass_sub(PyODictObject *od, PyObject *v, PyObject *w)
 static PyMappingMethods odict_as_mapping = {
     0,                                  /*mp_length*/
     0,                                  /*mp_subscript*/
-    (objobjargproc)odict_mp_ass_sub,    /*mp_ass_subscript*/
+    odict_mp_ass_sub,                   /*mp_ass_subscript*/
 };
 
 
@@ -948,28 +961,33 @@ OrderedDict_fromkeys_impl(PyTypeObject *type, PyObject *seq, PyObject *value)
     return _PyDict_FromKeys((PyObject *)type, seq, value);
 }
 
-/* __sizeof__() */
+/*[clinic input]
+@critical_section
+OrderedDict.__sizeof__ -> Py_ssize_t
+[clinic start generated code]*/
 
-/* OrderedDict.__sizeof__() does not have a docstring. */
-PyDoc_STRVAR(odict_sizeof__doc__, "");
-
-static PyObject *
-odict_sizeof(PyODictObject *od, PyObject *Py_UNUSED(ignored))
+static Py_ssize_t
+OrderedDict___sizeof___impl(PyODictObject *self)
+/*[clinic end generated code: output=1a8560db8cf83ac5 input=655e989ae24daa6a]*/
 {
-    Py_ssize_t res = _PyDict_SizeOf((PyDictObject *)od);
-    res += sizeof(_ODictNode *) * od->od_fast_nodes_size;  /* od_fast_nodes */
-    if (!_odict_EMPTY(od)) {
-        res += sizeof(_ODictNode) * PyODict_SIZE(od);  /* linked-list */
+    Py_ssize_t res = _PyDict_SizeOf_LockHeld((PyDictObject *)self);
+    res += sizeof(_ODictNode *) * self->od_fast_nodes_size;  /* od_fast_nodes */
+    if (!_odict_EMPTY(self)) {
+        res += sizeof(_ODictNode) * PyODict_SIZE(self);  /* linked-list */
     }
-    return PyLong_FromSsize_t(res);
+    return res;
 }
 
-/* __reduce__() */
+/*[clinic input]
+OrderedDict.__reduce__
+    self as od: self(type="PyODictObject *")
 
-PyDoc_STRVAR(odict_reduce__doc__, "Return state information for pickling");
+Return state information for pickling
+[clinic start generated code]*/
 
 static PyObject *
-odict_reduce(register PyODictObject *od, PyObject *Py_UNUSED(ignored))
+OrderedDict___reduce___impl(PyODictObject *od)
+/*[clinic end generated code: output=71eeb81f760a6a8e input=b0467c7ec400fe5e]*/
 {
     PyObject *state, *result = NULL;
     PyObject *items_iter, *items, *args = NULL;
@@ -1005,8 +1023,10 @@ Done:
 
 /* setdefault(): Skips __missing__() calls. */
 
+static int PyODict_SetItem_LockHeld(PyObject *self, PyObject *key, PyObject *value);
 
 /*[clinic input]
+@critical_section
 OrderedDict.setdefault
 
     key: object
@@ -1020,7 +1040,7 @@ Return the value for key if key is in the dictionary, else default.
 static PyObject *
 OrderedDict_setdefault_impl(PyODictObject *self, PyObject *key,
                             PyObject *default_value)
-/*[clinic end generated code: output=97537cb7c28464b6 input=38e098381c1efbc6]*/
+/*[clinic end generated code: output=97537cb7c28464b6 input=d7b93e92734f99b5]*/
 {
     PyObject *result = NULL;
 
@@ -1030,7 +1050,7 @@ OrderedDict_setdefault_impl(PyODictObject *self, PyObject *key,
             if (PyErr_Occurred())
                 return NULL;
             assert(_odict_find_node(self, key) == NULL);
-            if (PyODict_SetItem((PyObject *)self, key, default_value) >= 0) {
+            if (PyODict_SetItem_LockHeld((PyObject *)self, key, default_value) >= 0) {
                 result = Py_NewRef(default_value);
             }
         }
@@ -1060,16 +1080,15 @@ static PyObject *
 _odict_popkey_hash(PyObject *od, PyObject *key, PyObject *failobj,
                    Py_hash_t hash)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
+
     PyObject *value = NULL;
-
-    Py_BEGIN_CRITICAL_SECTION(od);
-
-    _ODictNode *node = _odict_find_node_hash((PyODictObject *)od, key, hash);
+    _ODictNode *node = _odict_find_node_hash(_PyODictObject_CAST(od), key, hash);
     if (node != NULL) {
         /* Pop the node first to avoid a possible dict resize (due to
            eval loop reentrancy) and complications due to hash collision
            resolution. */
-        int res = _odict_clear_node((PyODictObject *)od, node, key, hash);
+        int res = _odict_clear_node(_PyODictObject_CAST(od), node, key, hash);
         if (res < 0) {
             goto done;
         }
@@ -1088,7 +1107,6 @@ _odict_popkey_hash(PyObject *od, PyObject *key, PyObject *failobj,
             PyErr_SetObject(PyExc_KeyError, key);
         }
     }
-    Py_END_CRITICAL_SECTION();
 done:
 
     return value;
@@ -1096,6 +1114,7 @@ done:
 
 /* Skips __missing__() calls. */
 /*[clinic input]
+@critical_section
 OrderedDict.pop
 
     key: object
@@ -1110,7 +1129,7 @@ raise a KeyError.
 static PyObject *
 OrderedDict_pop_impl(PyODictObject *self, PyObject *key,
                      PyObject *default_value)
-/*[clinic end generated code: output=7a6447d104e7494b input=7efe36601007dff7]*/
+/*[clinic end generated code: output=7a6447d104e7494b input=a79988887b4a651f]*/
 {
     Py_hash_t hash = PyObject_Hash(key);
     if (hash == -1)
@@ -1122,6 +1141,7 @@ OrderedDict_pop_impl(PyODictObject *self, PyObject *key,
 /* popitem() */
 
 /*[clinic input]
+@critical_section
 OrderedDict.popitem
 
     last: bool = True
@@ -1133,7 +1153,7 @@ Pairs are returned in LIFO order if last is true or FIFO order if false.
 
 static PyObject *
 OrderedDict_popitem_impl(PyODictObject *self, int last)
-/*[clinic end generated code: output=98e7d986690d49eb input=d992ac5ee8305e1a]*/
+/*[clinic end generated code: output=98e7d986690d49eb input=8aafc7433e0a40e7]*/
 {
     PyObject *key, *value, *item = NULL;
     _ODictNode *node;
@@ -1162,6 +1182,9 @@ OrderedDict_popitem_impl(PyODictObject *self, int last)
 PyDoc_STRVAR(odict_keys__doc__, "");
 
 static PyObject * odictkeys_new(PyObject *od, PyObject *Py_UNUSED(ignored));  /* forward */
+static int
+_PyODict_SetItem_KnownHash_LockHeld(PyObject *od, PyObject *key, PyObject *value,
+                                    Py_hash_t hash); /* forward */
 
 /* values() */
 
@@ -1187,29 +1210,35 @@ static PyObject * mutablemapping_update(PyObject *, PyObject *, PyObject *);
 
 #define odict_update mutablemapping_update
 
-/* clear() */
+/*[clinic input]
+@critical_section
+OrderedDict.clear
 
-PyDoc_STRVAR(odict_clear__doc__,
-             "od.clear() -> None.  Remove all items from od.");
+Remove all items from ordered dict.
+[clinic start generated code]*/
 
 static PyObject *
-odict_clear(register PyODictObject *od, PyObject *Py_UNUSED(ignored))
+OrderedDict_clear_impl(PyODictObject *self)
+/*[clinic end generated code: output=a1a76d1322f556c5 input=08b12322e74c535c]*/
 {
-    PyDict_Clear((PyObject *)od);
-    _odict_clear_nodes(od);
+    _PyDict_Clear_LockHeld((PyObject *)self);
+    _odict_clear_nodes(self);
     Py_RETURN_NONE;
 }
 
 /* copy() */
 
-/* forward */
-static int _PyODict_SetItem_KnownHash(PyObject *, PyObject *, PyObject *,
-                                      Py_hash_t);
+/*[clinic input]
+@critical_section
+OrderedDict.copy
+    self as od: self
 
-PyDoc_STRVAR(odict_copy__doc__, "od.copy() -> a shallow copy of od");
+A shallow copy of ordered dict.
+[clinic start generated code]*/
 
 static PyObject *
-odict_copy(register PyODictObject *od, PyObject *Py_UNUSED(ignored))
+OrderedDict_copy_impl(PyObject *od)
+/*[clinic end generated code: output=9cdbe7394aecc576 input=e329951ae617ed48]*/
 {
     _ODictNode *node;
     PyObject *od_copy;
@@ -1230,8 +1259,8 @@ odict_copy(register PyODictObject *od, PyObject *Py_UNUSED(ignored))
                     PyErr_SetObject(PyExc_KeyError, key);
                 goto fail;
             }
-            if (_PyODict_SetItem_KnownHash((PyObject *)od_copy, key, value,
-                                           _odictnode_HASH(node)) != 0)
+            if (_PyODict_SetItem_KnownHash_LockHeld((PyObject *)od_copy, key, value,
+                                                    _odictnode_HASH(node)) != 0)
                 goto fail;
         }
     }
@@ -1269,8 +1298,9 @@ PyDoc_STRVAR(odict_reversed__doc__, "od.__reversed__() <==> reversed(od)");
 static PyObject * odictiter_new(PyODictObject *, int);
 
 static PyObject *
-odict_reversed(PyODictObject *od, PyObject *Py_UNUSED(ignored))
+odict_reversed(PyObject *op, PyObject *Py_UNUSED(ignored))
 {
+    PyODictObject *od = _PyODictObject_CAST(op);
     return odictiter_new(od, _odict_ITER_KEYS|_odict_ITER_REVERSED);
 }
 
@@ -1278,6 +1308,7 @@ odict_reversed(PyODictObject *od, PyObject *Py_UNUSED(ignored))
 /* move_to_end() */
 
 /*[clinic input]
+@critical_section
 OrderedDict.move_to_end
 
     key: object
@@ -1290,7 +1321,7 @@ Raise KeyError if the element does not exist.
 
 static PyObject *
 OrderedDict_move_to_end_impl(PyODictObject *self, PyObject *key, int last)
-/*[clinic end generated code: output=fafa4c5cc9b92f20 input=d6ceff7132a2fcd7]*/
+/*[clinic end generated code: output=fafa4c5cc9b92f20 input=09f8bc7053c0f6d4]*/
 {
     _ODictNode *node;
 
@@ -1331,10 +1362,8 @@ static PyMethodDef odict_methods[] = {
 
     /* overridden dict methods */
     ORDEREDDICT_FROMKEYS_METHODDEF
-    {"__sizeof__",      (PyCFunction)odict_sizeof,      METH_NOARGS,
-     odict_sizeof__doc__},
-    {"__reduce__",      (PyCFunction)odict_reduce,      METH_NOARGS,
-     odict_reduce__doc__},
+    ORDEREDDICT___SIZEOF___METHODDEF
+    ORDEREDDICT___REDUCE___METHODDEF
     ORDEREDDICT_SETDEFAULT_METHODDEF
     ORDEREDDICT_POP_METHODDEF
     ORDEREDDICT_POPITEM_METHODDEF
@@ -1346,13 +1375,10 @@ static PyMethodDef odict_methods[] = {
      odict_items__doc__},
     {"update",          _PyCFunction_CAST(odict_update), METH_VARARGS | METH_KEYWORDS,
      odict_update__doc__},
-    {"clear",           (PyCFunction)odict_clear,       METH_NOARGS,
-     odict_clear__doc__},
-    {"copy",            (PyCFunction)odict_copy,        METH_NOARGS,
-     odict_copy__doc__},
-
+     ORDEREDDICT_CLEAR_METHODDEF
+     ORDEREDDICT_COPY_METHODDEF
     /* new methods */
-    {"__reversed__",    (PyCFunction)odict_reversed,    METH_NOARGS,
+    {"__reversed__",    odict_reversed,    METH_NOARGS,
      odict_reversed__doc__},
     ORDEREDDICT_MOVE_TO_END_METHODDEF
 
@@ -1378,25 +1404,24 @@ static PyGetSetDef odict_getset[] = {
 /* tp_dealloc */
 
 static void
-odict_dealloc(PyODictObject *self)
+odict_dealloc(PyObject *op)
 {
+    PyODictObject *self = _PyODictObject_CAST(op);
     PyObject_GC_UnTrack(self);
-    Py_TRASHCAN_BEGIN(self, odict_dealloc)
 
     Py_XDECREF(self->od_inst_dict);
-    FT_CLEAR_WEAKREFS((PyObject*)self, self->od_weakreflist);
+    FT_CLEAR_WEAKREFS(op, self->od_weakreflist);
 
     _odict_clear_nodes(self);
     PyDict_Type.tp_dealloc((PyObject *)self);
-
-    Py_TRASHCAN_END
 }
 
 /* tp_repr */
 
 static PyObject *
-odict_repr(PyODictObject *self)
+odict_repr(PyObject *op)
 {
+    PyODictObject *self = _PyODictObject_CAST(op);
     int i;
     PyObject *result = NULL, *dcopy = NULL;
 
@@ -1431,8 +1456,9 @@ PyDoc_STRVAR(odict_doc,
 /* tp_traverse */
 
 static int
-odict_traverse(PyODictObject *od, visitproc visit, void *arg)
+odict_traverse(PyObject *op, visitproc visit, void *arg)
 {
+    PyODictObject *od = _PyODictObject_CAST(op);
     _ODictNode *node;
 
     Py_VISIT(od->od_inst_dict);
@@ -1445,10 +1471,12 @@ odict_traverse(PyODictObject *od, visitproc visit, void *arg)
 /* tp_clear */
 
 static int
-odict_tp_clear(PyODictObject *od)
+odict_tp_clear(PyObject *op)
 {
+    PyODictObject *od = _PyODictObject_CAST(op);
     Py_CLEAR(od->od_inst_dict);
-    PyDict_Clear((PyObject *)od);
+    // cannot use lock held variant as critical section is not held here
+    PyDict_Clear(op);
     _odict_clear_nodes(od);
     return 0;
 }
@@ -1456,7 +1484,7 @@ odict_tp_clear(PyODictObject *od)
 /* tp_richcompare */
 
 static PyObject *
-odict_richcompare(PyObject *v, PyObject *w, int op)
+odict_richcompare_lock_held(PyObject *v, PyObject *w, int op)
 {
     if (!PyODict_Check(v) || !PyDict_Check(w)) {
         Py_RETURN_NOTIMPLEMENTED;
@@ -1478,7 +1506,7 @@ odict_richcompare(PyObject *v, PyObject *w, int op)
         Py_DECREF(cmp);
 
         /* Try comparing odict keys. */
-        eq = _odict_keys_equal((PyODictObject *)v, (PyODictObject *)w);
+        eq = _odict_keys_equal(_PyODictObject_CAST(v), _PyODictObject_CAST(w));
         if (eq < 0)
             return NULL;
 
@@ -1489,12 +1517,22 @@ odict_richcompare(PyObject *v, PyObject *w, int op)
     }
 }
 
+static PyObject *
+odict_richcompare(PyObject *v, PyObject *w, int op)
+{
+    PyObject *res;
+    Py_BEGIN_CRITICAL_SECTION2(v, w);
+    res = odict_richcompare_lock_held(v, w, op);
+    Py_END_CRITICAL_SECTION2();
+    return res;
+}
+
 /* tp_iter */
 
 static PyObject *
-odict_iter(PyODictObject *od)
+odict_iter(PyObject *op)
 {
-    return odictiter_new(od, _odict_ITER_KEYS);
+    return odictiter_new(_PyODictObject_CAST(op), _odict_ITER_KEYS);
 }
 
 /* tp_init */
@@ -1530,12 +1568,12 @@ PyTypeObject PyODict_Type = {
     "collections.OrderedDict",                  /* tp_name */
     sizeof(PyODictObject),                      /* tp_basicsize */
     0,                                          /* tp_itemsize */
-    (destructor)odict_dealloc,                  /* tp_dealloc */
+    odict_dealloc,                              /* tp_dealloc */
     0,                                          /* tp_vectorcall_offset */
     0,                                          /* tp_getattr */
     0,                                          /* tp_setattr */
     0,                                          /* tp_as_async */
-    (reprfunc)odict_repr,                       /* tp_repr */
+    odict_repr,                                 /* tp_repr */
     &odict_as_number,                           /* tp_as_number */
     0,                                          /* tp_as_sequence */
     &odict_as_mapping,                          /* tp_as_mapping */
@@ -1547,11 +1585,11 @@ PyTypeObject PyODict_Type = {
     0,                                          /* tp_as_buffer */
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,/* tp_flags */
     odict_doc,                                  /* tp_doc */
-    (traverseproc)odict_traverse,               /* tp_traverse */
-    (inquiry)odict_tp_clear,                    /* tp_clear */
-    (richcmpfunc)odict_richcompare,             /* tp_richcompare */
+    odict_traverse,                             /* tp_traverse */
+    odict_tp_clear,                             /* tp_clear */
+    odict_richcompare,                          /* tp_richcompare */
     offsetof(PyODictObject, od_weakreflist),    /* tp_weaklistoffset */
-    (getiterfunc)odict_iter,                    /* tp_iter */
+    odict_iter,                                 /* tp_iter */
     0,                                          /* tp_iternext */
     odict_methods,                              /* tp_methods */
     0,                                          /* tp_members */
@@ -1561,7 +1599,7 @@ PyTypeObject PyODict_Type = {
     0,                                          /* tp_descr_get */
     0,                                          /* tp_descr_set */
     offsetof(PyODictObject, od_inst_dict),      /* tp_dictoffset */
-    (initproc)odict_init,                       /* tp_init */
+    odict_init,                                 /* tp_init */
     PyType_GenericAlloc,                        /* tp_alloc */
     0,                                          /* tp_new */
     0,                                          /* tp_free */
@@ -1579,12 +1617,13 @@ PyODict_New(void)
 }
 
 static int
-_PyODict_SetItem_KnownHash(PyObject *od, PyObject *key, PyObject *value,
-                           Py_hash_t hash)
+_PyODict_SetItem_KnownHash_LockHeld(PyObject *od, PyObject *key, PyObject *value,
+                                    Py_hash_t hash)
 {
-    int res = _PyDict_SetItem_KnownHash(od, key, value, hash);
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
+    int res = _PyDict_SetItem_KnownHash_LockHeld((PyDictObject *)od, key, value, hash);
     if (res == 0) {
-        res = _odict_add_new_node((PyODictObject *)od, key, hash);
+        res = _odict_add_new_node(_PyODictObject_CAST(od), key, hash);
         if (res < 0) {
             /* Revert setting the value on the dict */
             PyObject *exc = PyErr_GetRaisedException();
@@ -1595,29 +1634,52 @@ _PyODict_SetItem_KnownHash(PyObject *od, PyObject *key, PyObject *value,
     return res;
 }
 
+
+static int
+PyODict_SetItem_LockHeld(PyObject *od, PyObject *key, PyObject *value)
+{
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
+    Py_hash_t hash = PyObject_Hash(key);
+    if (hash == -1) {
+        return -1;
+    }
+    return _PyODict_SetItem_KnownHash_LockHeld(od, key, value, hash);
+}
+
 int
 PyODict_SetItem(PyObject *od, PyObject *key, PyObject *value)
 {
+    int res;
+    Py_BEGIN_CRITICAL_SECTION(od);
+    res = PyODict_SetItem_LockHeld(od, key, value);
+    Py_END_CRITICAL_SECTION();
+    return res;
+}
+
+int
+PyODict_DelItem_LockHeld(PyObject *od, PyObject *key)
+{
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(od);
+    int res;
     Py_hash_t hash = PyObject_Hash(key);
     if (hash == -1)
         return -1;
-    return _PyODict_SetItem_KnownHash(od, key, value, hash);
+    res = _odict_clear_node(_PyODictObject_CAST(od), NULL, key, hash);
+    if (res < 0)
+        return -1;
+    return _PyDict_DelItem_KnownHash_LockHeld(od, key, hash);
 }
 
 int
 PyODict_DelItem(PyObject *od, PyObject *key)
 {
     int res;
-    Py_hash_t hash = PyObject_Hash(key);
-    if (hash == -1)
-        return -1;
-    res = _odict_clear_node((PyODictObject *)od, NULL, key, hash);
-    if (res < 0)
-        return -1;
-    return _PyDict_DelItem_KnownHash(od, key, hash);
+    Py_BEGIN_CRITICAL_SECTION(od);
+    res = PyODict_DelItem_LockHeld(od, key);
+    Py_END_CRITICAL_SECTION();
+    return res;
 }
 
-
 /* -------------------------------------------
  * The OrderedDict views (keys/values/items)
  */
@@ -1633,8 +1695,9 @@ typedef struct {
 } odictiterobject;
 
 static void
-odictiter_dealloc(odictiterobject *di)
+odictiter_dealloc(PyObject *op)
 {
+    odictiterobject *di = (odictiterobject*)op;
     _PyObject_GC_UNTRACK(di);
     Py_XDECREF(di->di_odict);
     Py_XDECREF(di->di_current);
@@ -1645,8 +1708,9 @@ odictiter_dealloc(odictiterobject *di)
 }
 
 static int
-odictiter_traverse(odictiterobject *di, visitproc visit, void *arg)
+odictiter_traverse(PyObject *op, visitproc visit, void *arg)
 {
+    odictiterobject *di = (odictiterobject*)op;
     Py_VISIT(di->di_odict);
     Py_VISIT(di->di_current);  /* A key could be any type, not just str. */
     Py_VISIT(di->di_result);
@@ -1656,14 +1720,14 @@ odictiter_traverse(odictiterobject *di, visitproc visit, void *arg)
 /* In order to protect against modifications during iteration, we track
  * the current key instead of the current node. */
 static PyObject *
-odictiter_nextkey(odictiterobject *di)
+odictiter_nextkey_lock_held(odictiterobject *di)
 {
+    assert(di->di_odict != NULL);
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(di->di_odict);
     PyObject *key = NULL;
     _ODictNode *node;
     int reversed = di->kind & _odict_ITER_REVERSED;
 
-    if (di->di_odict == NULL)
-        return NULL;
     if (di->di_current == NULL)
         goto done;  /* We're already done. */
 
@@ -1708,9 +1772,25 @@ done:
     return key;
 }
 
+
 static PyObject *
-odictiter_iternext(odictiterobject *di)
+odictiter_nextkey(odictiterobject *di)
 {
+    _Py_CRITICAL_SECTION_ASSERT_OBJECT_LOCKED(di);
+    if (di->di_odict == NULL) {
+        return NULL;
+    }
+    PyObject *res;
+    Py_BEGIN_CRITICAL_SECTION(di->di_odict);
+    res = odictiter_nextkey_lock_held(di);
+    Py_END_CRITICAL_SECTION();
+    return res;
+}
+
+static PyObject *
+odictiter_iternext_lock_held(PyObject *op)
+{
+    odictiterobject *di = (odictiterobject*)op;
     PyObject *result, *value;
     PyObject *key = odictiter_nextkey(di);  /* new reference */
 
@@ -1722,14 +1802,12 @@ odictiter_iternext(odictiterobject *di)
         return key;
     }
 
-    value = PyODict_GetItem((PyObject *)di->di_odict, key);  /* borrowed */
-    if (value == NULL) {
+    if (PyDict_GetItemRef((PyObject *)di->di_odict, key, &value) != 1) {
         if (!PyErr_Occurred())
             PyErr_SetObject(PyExc_KeyError, key);
         Py_DECREF(key);
         goto done;
     }
-    Py_INCREF(value);
 
     /* Handle the values case. */
     if (!(di->kind & _odict_ITER_KEYS)) {
@@ -1740,7 +1818,7 @@ odictiter_iternext(odictiterobject *di)
     /* Handle the items case. */
     result = di->di_result;
 
-    if (Py_REFCNT(result) == 1) {
+    if (_PyObject_IsUniquelyReferenced(result)) {
         /* not in use so we can reuse it
          * (the common case during iteration) */
         Py_INCREF(result);
@@ -1748,9 +1826,7 @@ odictiter_iternext(odictiterobject *di)
         Py_DECREF(PyTuple_GET_ITEM(result, 1));  /* borrowed */
         // bpo-42536: The GC may have untracked this result tuple. Since we're
         // recycling it, make sure it's tracked again:
-        if (!_PyObject_GC_IS_TRACKED(result)) {
-            _PyObject_GC_TRACK(result);
-        }
+        _PyTuple_Recycle(result);
     }
     else {
         result = PyTuple_New(2);
@@ -1771,13 +1847,26 @@ done:
     return NULL;
 }
 
+static PyObject *
+odictiter_iternext(PyObject *op)
+{
+    PyObject *res;
+    Py_BEGIN_CRITICAL_SECTION(op);
+    res = odictiter_iternext_lock_held(op);
+    Py_END_CRITICAL_SECTION();
+    return res;
+}
+
+
 /* No need for tp_clear because odictiterobject is not mutable. */
 
 PyDoc_STRVAR(reduce_doc, "Return state information for pickling");
 
 static PyObject *
-odictiter_reduce(odictiterobject *di, PyObject *Py_UNUSED(ignored))
+odictiter_reduce(PyObject *op, PyObject *Py_UNUSED(ignored))
 {
+    odictiterobject *di = (odictiterobject*)op;
+
     /* copy the iterator state */
     odictiterobject tmp = *di;
     Py_XINCREF(tmp.di_odict);
@@ -1794,7 +1883,7 @@ odictiter_reduce(odictiterobject *di, PyObject *Py_UNUSED(ignored))
 }
 
 static PyMethodDef odictiter_methods[] = {
-    {"__reduce__", (PyCFunction)odictiter_reduce, METH_NOARGS, reduce_doc},
+    {"__reduce__", odictiter_reduce, METH_NOARGS, reduce_doc},
     {NULL,              NULL}           /* sentinel */
 };
 
@@ -1804,7 +1893,7 @@ PyTypeObject PyODictIter_Type = {
     sizeof(odictiterobject),                  /* tp_basicsize */
     0,                                        /* tp_itemsize */
     /* methods */
-    (destructor)odictiter_dealloc,            /* tp_dealloc */
+    odictiter_dealloc,                        /* tp_dealloc */
     0,                                        /* tp_vectorcall_offset */
     0,                                        /* tp_getattr */
     0,                                        /* tp_setattr */
@@ -1821,12 +1910,12 @@ PyTypeObject PyODictIter_Type = {
     0,                                        /* tp_as_buffer */
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,  /* tp_flags */
     0,                                        /* tp_doc */
-    (traverseproc)odictiter_traverse,         /* tp_traverse */
+    odictiter_traverse,                       /* tp_traverse */
     0,                                        /* tp_clear */
     0,                                        /* tp_richcompare */
     0,                                        /* tp_weaklistoffset */
     PyObject_SelfIter,                        /* tp_iter */
-    (iternextfunc)odictiter_iternext,         /* tp_iternext */
+    odictiter_iternext,                       /* tp_iternext */
     odictiter_methods,                        /* tp_methods */
     0,
 };
@@ -1867,27 +1956,29 @@ odictiter_new(PyODictObject *od, int kind)
 /* keys() */
 
 static PyObject *
-odictkeys_iter(_PyDictViewObject *dv)
+odictkeys_iter(PyObject *op)
 {
+    _PyDictViewObject *dv = (_PyDictViewObject*)op;
     if (dv->dv_dict == NULL) {
         Py_RETURN_NONE;
     }
-    return odictiter_new((PyODictObject *)dv->dv_dict,
+    return odictiter_new(_PyODictObject_CAST(dv->dv_dict),
             _odict_ITER_KEYS);
 }
 
 static PyObject *
-odictkeys_reversed(_PyDictViewObject *dv, PyObject *Py_UNUSED(ignored))
+odictkeys_reversed(PyObject *op, PyObject *Py_UNUSED(ignored))
 {
+    _PyDictViewObject *dv = (_PyDictViewObject*)op;
     if (dv->dv_dict == NULL) {
         Py_RETURN_NONE;
     }
-    return odictiter_new((PyODictObject *)dv->dv_dict,
+    return odictiter_new(_PyODictObject_CAST(dv->dv_dict),
             _odict_ITER_KEYS|_odict_ITER_REVERSED);
 }
 
 static PyMethodDef odictkeys_methods[] = {
-    {"__reversed__", (PyCFunction)odictkeys_reversed, METH_NOARGS, NULL},
+    {"__reversed__", odictkeys_reversed, METH_NOARGS, NULL},
     {NULL,          NULL}           /* sentinel */
 };
 
@@ -1917,7 +2008,7 @@ PyTypeObject PyODictKeys_Type = {
     0,                                        /* tp_clear */
     0,                                        /* tp_richcompare */
     0,                                        /* tp_weaklistoffset */
-    (getiterfunc)odictkeys_iter,              /* tp_iter */
+    odictkeys_iter,                           /* tp_iter */
     0,                                        /* tp_iternext */
     odictkeys_methods,                        /* tp_methods */
     0,                                        /* tp_members */
@@ -1934,27 +2025,29 @@ odictkeys_new(PyObject *od, PyObject *Py_UNUSED(ignored))
 /* items() */
 
 static PyObject *
-odictitems_iter(_PyDictViewObject *dv)
+odictitems_iter(PyObject *op)
 {
+    _PyDictViewObject *dv = (_PyDictViewObject*)op;
     if (dv->dv_dict == NULL) {
         Py_RETURN_NONE;
     }
-    return odictiter_new((PyODictObject *)dv->dv_dict,
+    return odictiter_new(_PyODictObject_CAST(dv->dv_dict),
             _odict_ITER_KEYS|_odict_ITER_VALUES);
 }
 
 static PyObject *
-odictitems_reversed(_PyDictViewObject *dv, PyObject *Py_UNUSED(ignored))
+odictitems_reversed(PyObject *op, PyObject *Py_UNUSED(ignored))
 {
+    _PyDictViewObject *dv = (_PyDictViewObject*)op;
     if (dv->dv_dict == NULL) {
         Py_RETURN_NONE;
     }
-    return odictiter_new((PyODictObject *)dv->dv_dict,
+    return odictiter_new(_PyODictObject_CAST(dv->dv_dict),
             _odict_ITER_KEYS|_odict_ITER_VALUES|_odict_ITER_REVERSED);
 }
 
 static PyMethodDef odictitems_methods[] = {
-    {"__reversed__", (PyCFunction)odictitems_reversed, METH_NOARGS, NULL},
+    {"__reversed__", odictitems_reversed, METH_NOARGS, NULL},
     {NULL,          NULL}           /* sentinel */
 };
 
@@ -1984,7 +2077,7 @@ PyTypeObject PyODictItems_Type = {
     0,                                        /* tp_clear */
     0,                                        /* tp_richcompare */
     0,                                        /* tp_weaklistoffset */
-    (getiterfunc)odictitems_iter,             /* tp_iter */
+    odictitems_iter,                          /* tp_iter */
     0,                                        /* tp_iternext */
     odictitems_methods,                       /* tp_methods */
     0,                                        /* tp_members */
@@ -2001,27 +2094,29 @@ odictitems_new(PyObject *od, PyObject *Py_UNUSED(ignored))
 /* values() */
 
 static PyObject *
-odictvalues_iter(_PyDictViewObject *dv)
+odictvalues_iter(PyObject *op)
 {
+    _PyDictViewObject *dv = (_PyDictViewObject*)op;
     if (dv->dv_dict == NULL) {
         Py_RETURN_NONE;
     }
-    return odictiter_new((PyODictObject *)dv->dv_dict,
+    return odictiter_new(_PyODictObject_CAST(dv->dv_dict),
             _odict_ITER_VALUES);
 }
 
 static PyObject *
-odictvalues_reversed(_PyDictViewObject *dv, PyObject *Py_UNUSED(ignored))
+odictvalues_reversed(PyObject *op, PyObject *Py_UNUSED(ignored))
 {
+    _PyDictViewObject *dv = (_PyDictViewObject*)op;
     if (dv->dv_dict == NULL) {
         Py_RETURN_NONE;
     }
-    return odictiter_new((PyODictObject *)dv->dv_dict,
+    return odictiter_new(_PyODictObject_CAST(dv->dv_dict),
             _odict_ITER_VALUES|_odict_ITER_REVERSED);
 }
 
 static PyMethodDef odictvalues_methods[] = {
-    {"__reversed__", (PyCFunction)odictvalues_reversed, METH_NOARGS, NULL},
+    {"__reversed__", odictvalues_reversed, METH_NOARGS, NULL},
     {NULL,          NULL}           /* sentinel */
 };
 
@@ -2051,7 +2146,7 @@ PyTypeObject PyODictValues_Type = {
     0,                                        /* tp_clear */
     0,                                        /* tp_richcompare */
     0,                                        /* tp_weaklistoffset */
-    (getiterfunc)odictvalues_iter,            /* tp_iter */
+    odictvalues_iter,                         /* tp_iter */
     0,                                        /* tp_iternext */
     odictvalues_methods,                      /* tp_methods */
     0,                                        /* tp_members */
