@@ -106,7 +106,10 @@ c_parser.add_argument(
     help="Suppress code emission for rule actions",
 )
 
-python_parser = subparsers.add_parser("python", help="Generate Python code")
+python_parser = subparsers.add_parser(
+    "python",
+    help="Generate Python code, needs grammar definition with Python actions",
+)
 python_parser.set_defaults(func=generate_python_code)
 python_parser.add_argument("grammar_filename", help="Grammar description")
 python_parser.add_argument(
